searchNodes=[{"doc":"The Dagex library is used to allow your business entities to participate in directed, acyclic graphs backed by PostgreSQL's ltree extenstion . N.B. The callbacks defined in this module are automatically defined for you on your Ecto models when you use Dagex inside them. Installation See the instructions in the project's README to perform initial setup of the required database tables and functions. Adding and manipulating a DAG for your business entity Let's say your application deals with classifying different types of animals. You may wish to model a set of animals as follows (taken from a post by Kemal Erdogan ): graph TD Animal --&gt; Pet Animal --&gt; Livestock Pet --&gt; Cat Pet --&gt; Dog Pet --&gt; Sheep Dog --&gt; Doberman Dog --&gt; Bulldog Livestock --&gt; Dog Livestock --&gt; Sheep Livestock --&gt; Cow Assuming you have configured Dagex as per the README , we can go ahead and set up the Ecto model needed. First, run mix ecto.gen.migration add_animal_types and then edit the resulting migration file: defmodule DagexTest.Repo.Migrations.AddAnimalTypes do def change do create table ( &quot;animal_types&quot; ) do add :name , :string , null : false timestamps ( ) end create index ( &quot;animal_types&quot; , :name , unique : true ) Dagex.Migrations . setup_node_type ( &quot;animal_types&quot; , &quot;3.0.0&quot; ) end end Run migrations with mix ecto.migrate and then create your model: defmodule DagexTest.AnimalType do use Ecto.Schema use Dagex schema &quot;animal_types&quot; do field :name , :string timestamps ( ) end end Now we can create the animal types and specify their associations in the DAG, and make queries about the graph itself: iex&gt; alias DagexTest . { AnimalType , Repo } iex&gt; iex&gt; # allows us to compare the contents of two lists independent of order iex&gt; import Assertions , only : [ assert_lists_equal : 2 ] iex&gt; iex&gt; # Add a couple of nodes iex&gt; { :ok , animal } = % AnimalType { name : &quot;Animal&quot; } |&gt; Repo . insert ( ) iex&gt; { :ok , pet } = % AnimalType { name : &quot;Pet&quot; } |&gt; Repo . insert ( ) iex&gt; iex&gt; # Create an association between the nodes iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( animal , pet ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; # Add the remaining nodes and create the associations as per the graph iex&gt; iex&gt; { :ok , livestock } = % AnimalType { name : &quot;Livestock&quot; } |&gt; Repo . insert ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( animal , livestock ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; { :ok , cat } = % AnimalType { name : &quot;Cat&quot; } |&gt; Repo . insert ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( pet , cat ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; # Note that a node may have multiple parents iex&gt; { :ok , dog } = % AnimalType { name : &quot;Dog&quot; } |&gt; Repo . insert ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( pet , dog ) |&gt; Repo . dagex_update ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( livestock , dog ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; { :ok , sheep } = % AnimalType { name : &quot;Sheep&quot; } |&gt; Repo . insert ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( pet , sheep ) |&gt; Repo . dagex_update ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( livestock , sheep ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; { :ok , cow } = % AnimalType { name : &quot;Cow&quot; } |&gt; Repo . insert ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( livestock , cow ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; { :ok , doberman } = % AnimalType { name : &quot;Doberman&quot; } |&gt; Repo . insert ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( dog , doberman ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; { :ok , bulldog } = % AnimalType { name : &quot;Bulldog&quot; } |&gt; Repo . insert ( ) iex&gt; { :edge_created , _edge } = AnimalType . create_edge ( dog , bulldog ) |&gt; Repo . dagex_update ( ) iex&gt; iex&gt; # we can get the direct children of a node: iex&gt; children = AnimalType . children ( animal ) |&gt; Repo . all ( ) iex&gt; assert_lists_equal ( children , [ pet , livestock ] ) iex&gt; iex&gt; # we can get the direct parents of a node iex&gt; parents = AnimalType . parents ( dog ) |&gt; Repo . all ( ) iex&gt; assert_lists_equal ( parents , [ pet , livestock ] ) iex&gt; iex&gt; # we can get all of the descendants of a node iex&gt; descendants = AnimalType . descendants ( pet ) |&gt; Repo . all ( ) iex&gt; assert_lists_equal ( descendants , [ cat , dog , sheep , doberman , bulldog ] ) iex&gt; iex&gt; # we can also get the descendants including the node itself iex&gt; descendants = AnimalType . with_descendants ( pet ) |&gt; Repo . all ( ) iex&gt; assert_lists_equal ( descendants , [ pet , cat , dog , sheep , doberman , bulldog ] ) iex&gt; iex&gt; # we can get all ancestors of a node iex&gt; ancestors = AnimalType . ancestors ( bulldog ) |&gt; Repo . all ( ) iex&gt; assert_lists_equal ( ancestors , [ dog , pet , livestock , animal ] ) iex&gt; iex&gt; # we can also get the ancestors including the node itself iex&gt; ancestors = AnimalType . with_ancestors ( bulldog ) |&gt; Repo . all ( ) iex&gt; assert_lists_equal ( ancestors , [ bulldog , dog , pet , livestock , animal ] ) iex&gt; iex&gt; # we can determine if node A precedes (i.e. is an ancestor of) node B iex&gt; true = AnimalType . precedes? ( pet , bulldog ) |&gt; Repo . exists? ( ) iex&gt; false = AnimalType . precedes? ( sheep , bulldog ) |&gt; Repo . exists? ( ) iex&gt; iex&gt; # and we can determine if node A succeeds (i.e. is a descendant of) node B iex&gt; true = AnimalType . succeeds? ( bulldog , pet ) |&gt; Repo . exists? ( ) iex&gt; false = AnimalType . succeeds? ( bulldog , sheep ) |&gt; Repo . exists? ( ) iex&gt; iex&gt; # we can also get the possible paths between two nodes iex&gt; paths = AnimalType . all_paths ( animal , sheep ) |&gt; Repo . dagex_paths ( ) iex&gt; assert_lists_equal ( paths , [ [ animal , pet , sheep ] , [ animal , livestock , sheep ] ] ) iex&gt; iex&gt; # if we remove an edge iex&gt; { :edge_removed , _edge } = AnimalType . remove_edge ( livestock , dog ) |&gt; Repo . dagex_update ( ) iex&gt; # then iex&gt; false = AnimalType . succeeds? ( bulldog , livestock ) |&gt; Repo . exists? ( ) iex&gt; false = AnimalType . precedes? ( livestock , bulldog ) |&gt; Repo . exists? ( ) iex&gt; iex&gt; # and if we remove a node entirely iex&gt; Repo . delete! ( livestock ) iex&gt; # then iex&gt; [ ] = AnimalType . ancestors ( cow ) |&gt; Repo . all ( ) iex&gt; assert_lists_equal ( [ dog , pet , animal ] , AnimalType . ancestors ( bulldog ) |&gt; Repo . all ( ) )","ref":"Dagex.html","title":"Dagex","type":"behaviour"},{"doc":"Returns a query that can be passed to your application's Ecto repository's Dagex.Repo.dagex_paths/1 function in order to retrieve a list of the possible paths between two nodes. Each path is itself a list starting with the ancestor node, ending with the descendant node, and including each node in the path between the two. Returns an empty list if no path exists between the two nodes.","ref":"Dagex.html#c:all_paths/2","title":"Dagex.all_paths/2","type":"callback"},{"doc":"Returns a query that can be passed to your application's Ecto repository to retrieve a list of entities that are ancestors of the specified child entity.","ref":"Dagex.html#c:ancestors/1","title":"Dagex.ancestors/1","type":"callback"},{"doc":"Returns a query that can be passed to your application's Ecto repository to retrieve a list of entities that are children of the specified parent entity.","ref":"Dagex.html#c:children/1","title":"Dagex.children/1","type":"callback"},{"doc":"Returns a Dagex.Operations.CreateEdge struct to be passed to Dagex.Repo.dagex_update/1 that will attempt to create a new edge in the implementing module's associated DAG.","ref":"Dagex.html#c:create_edge/2","title":"Dagex.create_edge/2","type":"callback"},{"doc":"Returns a query that can be passed to your application's Ecto repository to retrieve a list of entities that are descendants of the specified parent entity.","ref":"Dagex.html#c:descendants/1","title":"Dagex.descendants/1","type":"callback"},{"doc":"Returns a query that can be passed to your application's Ecto repository to retrieve a list of entities that are parents of the specified child entity.","ref":"Dagex.html#c:parents/1","title":"Dagex.parents/1","type":"callback"},{"doc":"Returns a query that selects ancestor only if ancestor is an ancestor of descendant.","ref":"Dagex.html#c:precedes?/2","title":"Dagex.precedes?/2","type":"callback"},{"doc":"Returns a Dagex.Operations.RemoveEdge struct to be passed to Dagex.Repo.dagex_update/1 that will attempt to remove the specified edge from the implementing module's associated DAG.","ref":"Dagex.html#c:remove_edge/2","title":"Dagex.remove_edge/2","type":"callback"},{"doc":"Returns a query that can be passed to your application's Ecto repository to retrieve a list of entities of the type defined in this module that are at the top level of the DAG (i.e. have no other parents.)","ref":"Dagex.html#c:roots/0","title":"Dagex.roots/0","type":"callback"},{"doc":"Returns a query that selects descendant only if descendant is a descendant of ancestor.","ref":"Dagex.html#c:succeeds?/2","title":"Dagex.succeeds?/2","type":"callback"},{"doc":"Returns a query that can be passed to your application's Ecto repository to retrieve a list of entities that are ancestors of the specified child entity as well as the child entity itself.","ref":"Dagex.html#c:with_ancestors/1","title":"Dagex.with_ancestors/1","type":"callback"},{"doc":"Returns a query that can be passed to your application's Ecto repository to retrieve a list of entities that are descendants of the specified parent entity along with the parent entity itself.","ref":"Dagex.html#c:with_descendants/1","title":"Dagex.with_descendants/1","type":"callback"},{"doc":"Migrations create and modify the database tables Dagex needs to function. Usage To use migrations in your application you'll need to generate an Ecto.Migration that wraps calls to Dagex.Migrations : mix ecto.gen.migration add_dagex Open the generated migration in your editor and call the up and down functions on Dagex.Migrations : defmodule MyApp.Repo.Migrations.AddDagex do use Ecto.Migration def up , do : Dagex.Migrations . up ( ) def down , do : Dagex.Migrations . down ( ) end This will run all of Dagex's versioned migrations for your database. Migrations between versions are idempotent. As new versions are released you may need to run additional migrations. Now, run the migration to create the table: mix ecto.migrate","ref":"Dagex.Migrations.html","title":"Dagex.Migrations","type":"module"},{"doc":"Run the down changes for all migrations between the current version and the initial version. Example Run all migrations from current version down to the first: Dagex.Migrations . down ( ) Run migrations down to a specified version: Dagex.Migrations . down ( version : 5 )","ref":"Dagex.Migrations.html#down/1","title":"Dagex.Migrations.down/1","type":"function"},{"doc":"","ref":"Dagex.Migrations.html#setup_node_type/2","title":"Dagex.Migrations.setup_node_type/2","type":"macro"},{"doc":"Run the up changes for all migrations between the initial version and the current version. Example Run all migrations up to the current version: Dagex.Migrations . up ( ) Run migrations up to a specified version: Dagex.Migrations . up ( version : 2 )","ref":"Dagex.Migrations.html#up/1","title":"Dagex.Migrations.up/1","type":"function"},{"doc":"Represents a database operation to create a new edge between two nodes. See Dagex.create_edge/2","ref":"Dagex.Operations.CreateEdge.html","title":"Dagex.Operations.CreateEdge","type":"module"},{"doc":"","ref":"Dagex.Operations.CreateEdge.html#t:result/0","title":"Dagex.Operations.CreateEdge.result/0","type":"type"},{"doc":"","ref":"Dagex.Operations.CreateEdge.html#t:t/0","title":"Dagex.Operations.CreateEdge.t/0","type":"type"},{"doc":"Represents a database operation to remove an edge between two nodes. See Dagex.remove_edge/2","ref":"Dagex.Operations.RemoveEdge.html","title":"Dagex.Operations.RemoveEdge","type":"module"},{"doc":"","ref":"Dagex.Operations.RemoveEdge.html#t:result/0","title":"Dagex.Operations.RemoveEdge.result/0","type":"type"},{"doc":"","ref":"Dagex.Operations.RemoveEdge.html#t:t/0","title":"Dagex.Operations.RemoveEdge.t/0","type":"type"},{"doc":"Adds Dagex-specific functionality to your application's Ecto.Repo module. defmodule MyApp.Repo do use Ecto.Repo , otp_app : :my_app , adapter : Ecto.Adapters.Postgres use Dagex.Repo end","ref":"Dagex.Repo.html","title":"Dagex.Repo","type":"behaviour"},{"doc":"Executes the query generated by Dagex.all_paths/2 and processes the result into a list of paths where each path is a list of the nodes between (and including) the ancestor and the descendant nodes.","ref":"Dagex.Repo.html#c:dagex_paths/1","title":"Dagex.Repo.dagex_paths/1","type":"callback"},{"doc":"Executes a Dagex repo operation such as Dagex.Operations.CreateEdge and processes the result.","ref":"Dagex.Repo.html#c:dagex_update/1","title":"Dagex.Repo.dagex_update/1","type":"callback"},{"doc":"Dagex provides tools to create directed, acyclic graphs for your business entities. It relies on the PostgreSQL ltree extension and Ecto . See the module documentation for Dagex for usage examples.","ref":"readme.html","title":"Dagex","type":"extras"},{"doc":"1. Add dependency If available in Hex , the package can be installed by adding dagex to your list of dependencies in mix.exs : def deps do [ { :dagex , &quot;~&gt; 3.0&quot; } ] end 2. Add Dagex functionality to your Ecto repository: defmodule MyApp.Repo do use Ecto.Repo , otp_app : :my_app , adapter : Ecto.Adapters.Postgres use Dagex.Repo end 3. Generate a migration to add the necessary database tables and functions Run mix ecto.gen.migration add_dagex_to_project then edit the resulting migration file: defmodule MyApp.Repo.Migrations.AddDagexToProject do use Ecto.Migration def up do Dagex.Migrations . up ( ) end def down do Dagex.Migrations . down ( ) end end and then migrate your database with mix ecto.migrate . 4. Use migrations to create db tables for the business entities that participate in a DAG: Run mix ecto.gen.migration add_organizations then edit the resulting migration file: defmodule MyApp.Repo.Migrations.AddOrganizations do require Dagex.Migrations use Ecto.Migration def change do create table ( &quot;organizations&quot; ) do add :name , :string , null : false timestamps ( ) end # Adds triggers to the &quot;organizations&quot; table to maintain the associated DAG as # records are added/removed. Dagex.Migrations . setup_node_type ( &quot;organizations&quot; , &quot;3.0.0&quot; ) end end 5. Create your Ecto-backed entity and add Dagex functionality: N.B. You can use any column type that Postgresql can convert to text as the entity's primary key, however, if your primary key field is a string, you MUST NOT allow the value &quot;*&quot; as this is reserved internally for the supremum of the graph. You will see an Ecto.ConstrainError on the dagex_reserved_supremum_id constraint if you attempt to insert such a record. defmodule MyApp.Organization do use Dagex use Ecto.Schema schema &quot;organizations&quot; do field :name , :string end end See the module documentation for Dagex for information on the functionality that is added to your module when calling use Dagex .","ref":"readme.html#installation","title":"Dagex - Installation","type":"extras"},{"doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS APPENDIX: How to apply the Apache License to your work. To apply the Apache License to your work, attach the following boilerplate notice, with the fields enclosed by brackets &quot;[]&quot; replaced with your own identifying information. (Don't include the brackets!) The text should be enclosed in the appropriate comment syntax for the file format. We also recommend that a file or class name and description of purpose be included on the same &quot;printed page&quot; as the copyright notice for easier identification within third-party archives. Copyright 2022 John Wilger Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"license.html","title":"LICENSE","type":"extras"}]